@@ -1,7 +1,7 @@
-/* ttywatch.c -- log output from TTYs, based on robin.c from the book
+/* ttywatch.c -- log output from TTYs, once loosely based on robin.c from
  * Linux Application Development, by Michael K. Johnson and Erik W. Troan
  *
- * Copyright  2000 Michael K. Johnson <johnsonm@redhat.com>
+ * Copyright  2000, 2001 Michael K. Johnson <johnsonm@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -32,9 +32,11 @@
 #include <string.h>             /* for strerror() */
 #include <unistd.h>
 
-#include <glib.h>
-
 #include "ttywatch.h"
+#include "logfile.h"
+#include "errors.h"
+#include "telnet.h"
+#include "socket.h"
 
 /* configuration */
 static char *logpath;
@@ -53,11 +55,17 @@ static struct poptOption optionsTable[] = {
     { "name", 'n', POPT_ARG_STRING, &chararg, 'n',
       "name of machine to monitor", "machinename" },
     { "port", 'p', POPT_ARG_STRING, &chararg, 'p',
-      "pathname for port for current machine", "/dev/<foo>" },
+      "pathname for port for current machinename", "/dev/<foo>" },
+    { "ipport", 'i', POPT_ARG_STRING, &chararg, 'i',
+      "port number on which to talk to this device", "3000" },
+    { "ipaddr", 'I', POPT_ARG_STRING, &chararg, 'I',
+      "IP address to which to bind", "10.0.0.1" },
     { "bps",  'b', POPT_ARG_INT, &intarg, 'b',
       "signaling rate for current maching in bps", "<BPS>" },
     { "logpath", 'l', POPT_ARG_STRING, &logpath, 0,
       "directory path for logfiles", "path" },
+    { "logspew", 'L', POPT_ARG_INT, &intarg, 'L',
+      "characters of logfile data to send", "4000" },
     { "module", 'm', POPT_ARG_STRING, &modname, 'm',
       "path to module for logging this port", "path" },
     { "daemon", 'd', POPT_ARG_NONE, &do_daemon, 0,
@@ -69,24 +77,8 @@ static struct poptOption optionsTable[] = {
 static GSList *machines;
 static GHashTable *machine_names;
 static GHashTable *machine_fds;
-
-void warn(char *warning, machine *m) {
-	char * message;
-	int size;
-
-	if (!m) abort();
-
-	message = alloca(strlen(warning) + strlen(m->name) + 15);
-	size = sprintf(message, "ttywatch[%s]: %s\n", m->name, warning);
-	write(STDERR_FILENO, message, size);
-	if (m->log_fd >= 0) write(m->log_fd, message, size);
-}
-
-/* eeeeeek! */
-void die(char *message) {
-	write(STDERR_FILENO, message, strlen(message));
-	exit (1);
-}
+static GHashTable *accept_fds;
+static GHashTable *net_fds;
 
 /* get machine* for machine name, creating a new machine* if necessary */
 machine *get_machine(char *mach) {
@@ -100,14 +92,15 @@ machine *get_machine(char *mach) {
 
 	m->name = strdup(mach);
 	m->speed = B115200;
+	m->logspew = 80*50; /* up to at least two screen pages */
 
 	/* defaults to devpath = /dev/<name> for symlinks */
 	m->devpath = malloc(strlen(mach)+6);
 	strcpy(m->devpath, "/dev/");
 	strcat(m->devpath, mach);
 
-	g_hash_table_insert(machine_names, mach, m);
-	machines = g_slist_append(machines, m);
+	g_hash_table_insert(machine_names, m->name, m);
+	machines = g_slist_prepend(machines, m);
 
 	return m;
 }
@@ -152,28 +145,17 @@ void delete_pidfile(void) {
 	}
 }
 
-void open_one_logfile(machine *m) {
-	char *logname;
-
-	/* no logfile if a module is being used */
-	if (m->process_line) return;
-
-	logname = alloca(strlen(logpath)+strlen(m->name)+6);
-	sprintf(logname, "%s/%s.log", logpath, m->name);
-	m->log_fd = open(logname, O_RDWR|O_CREAT|O_APPEND, 0666);
-	if (m->log_fd < 0)
-		warn(strerror(errno), m);
-}
-
-/* open and set termios settings for a single port, storing original settings */
-void setup_one_termios(gpointer data, gpointer ignore) {
+/* open and set termios settings for a single port, storing original settings
+ * also set up ip port listening if requested
+ */
+void setup_one_device(gpointer data, gpointer ignore) {
 	machine *m = data;
 
 	open_one_logfile(m);
 
-	m->dev_fd = open(m->devpath, O_RDONLY|O_NOCTTY);
+	m->dev_fd = open(m->devpath, O_RDWR|O_NOCTTY);
 	if (m->dev_fd < 0) {
-		warn(strerror(errno), m);
+		warn(m, strerror(errno));
 		g_hash_table_remove(machine_names, m->name);
 		machines = g_slist_remove(machines, m);
 		free(m->devpath);
@@ -204,60 +186,108 @@ void setup_one_termios(gpointer data, gpointer ignore) {
 
 	/* Now set the modified termios settings */
 	tcsetattr(m->dev_fd, TCSANOW, &m->ts);
+
+	/* Are we supposed to export this port via the network? */
+	if (m->ipport) {
+	    m->accept_fd = socket_listen(m->ipport, m->ipaddr);
+	    if (m->accept_fd) {
+		g_hash_table_insert(accept_fds, &m->accept_fd, m);
+	    }
+	}
 }
 
 /* set termios settings for all ports */
-void setup_termios(void) {
-	g_slist_foreach(machines, setup_one_termios, NULL);
+void setup_devices(void) {
+	g_slist_foreach(machines, setup_one_device, NULL);
 }
 
-/* restore termios settings for a single port, then close fds */
-void cleanup_one_termios(gpointer data, gpointer ignore) {
+/* restore termios settings for a single port, then close all fds */
+void cleanup_one_device(gpointer data, gpointer ignore) {
 	machine *m = data;
 
 	tcsetattr(m->dev_fd, TCSANOW, &m->ots);
 	close(m->dev_fd);
 	/* no logfile if a module is being used */
 	if (m->log_fd) close(m->log_fd);
+	if (m->accept_fd) close(m->accept_fd);
+	if (m->net_conns) {
+	    GSList *l;
+	    for (l = m->net_conns; l; l = g_slist_next(l)) {
+		net_conn *nc = l->data;
+
+		close(nc->fd);
+		g_hash_table_remove(net_fds, &nc->fd);
+		free(nc);
+	    }
+	    g_slist_free(m->net_conns);
+	}
+}
+
+void
+reopen_logfiles(int signo) {
+    g_slist_foreach(machines, reopen_one_logfile, NULL);
 }
 
 /* restore all original terminal settings on exit */
-void cleanup_termios_and_exit(int signal) {
-	g_slist_foreach(machines, cleanup_one_termios, NULL);
+void cleanup_devices_and_exit(int signal) {
+	g_slist_foreach(machines, cleanup_one_device, NULL);
 	delete_pidfile();
 	exit(0);
 }
 
-void reopen_one_logfile(gpointer data, gpointer ignore) {
-	machine *m = data;
-
-	/* no logfile if a module is being used */
-	if (m->log_fd) close(m->log_fd);
-	open_one_logfile(m);
-}
-
-void reopen_logfiles(int signo) {
-	g_slist_foreach(machines, reopen_one_logfile, NULL);
-}
-
 struct pollfd * prepare_poll(int *nfds /*OUT*/) {
 	struct pollfd *p;
-	GSList *l;
+	GSList *l, *n;
 	machine *m;
 	int i;
 
-	*nfds = g_slist_length(machines);
+	*nfds=0;
+
+	for (l = machines; l; l = g_slist_next(l)) {
+	    m = l->data;
+	    (*nfds)++; /* one for the port */
+	    if (m->accept_fd)
+		(*nfds)++; /* one for the accept fd */
+	    if (m->net_conns)
+		for (n = m->net_conns; n; n = g_slist_next(n))
+		    (*nfds)++; /* one for each listening client */
+	}
 	p = calloc(sizeof(struct pollfd), *nfds);
 
-	for (l = machines, i=0; l; l = g_slist_next(l), i++) {
-		m = l->data;
-		p[i].fd = m->dev_fd;
-		p[i].events = POLLIN;
+	for (l = machines, i=0; l; l = g_slist_next(l)) {
+	    m = l->data;
+	    p[i].fd = m->dev_fd;
+	    p[i++].events = POLLIN;
+	    if (m->accept_fd) {
+		p[i].fd = m->accept_fd;
+		p[i++].events = POLLIN;
+	    }
+	    if (m->net_conns) {
+		for (n = m->net_conns; n; n = g_slist_next(n)) {
+		    net_conn *nc = n->data;
+
+		    p[i].fd = nc->fd;
+		    p[i++].events = POLLIN;
+		}
+	    }
 	}
 
 	return p;
 }
 
+/* A network client has gone away */
+void
+remove_client (net_conn *nc, struct pollfd *ufds, int *nfds) {
+	machine *m;
+
+	close(nc->fd);
+	m = nc->m;
+	m->net_conns = g_slist_remove(m->net_conns, nc);
+	g_hash_table_remove(net_fds, nc);
+	free(nc);
+	ufds = prepare_poll(nfds);
+}
+
 
 machine *parse_argument (char c, machine *m) {
 	int	conf_fd = 0;
@@ -267,6 +297,16 @@ machine *parse_argument (char c, machine *m) {
 	case 'n':
 		m = get_machine(chararg);
 		break;
+	case 'i':
+		if (!m) die ("--name or -n must come before --ipport or -i\n");
+		if (m->ipport) free(m->ipport);
+		m->ipport = strdup(chararg);
+		break;
+	case 'I':
+		if (!m) die ("--name or -n must come before --ipaddr or -I\n");
+		if (m->ipaddr) free(m->ipaddr);
+		m->ipaddr = strdup(chararg);
+		break;
 	case 'p':
 		if (!m) die ("--name or -n must come before --port or -p\n");
 		if (m->devpath) free(m->devpath);
@@ -276,6 +316,10 @@ machine *parse_argument (char c, machine *m) {
 		if (!m) die ("--name or -n must come before --bps or -b\n");
 		m->speed = symbolic_speed(intarg);
 		break;
+	case 'L':
+		if (!m) die ("--name or -n must come before --logspew or -L\n");
+		m->logspew = intarg;
+		break;
 	case 'm':
 		if (!m) die ("--name or -n must come before --module or -m\n");
 		if (m->modhandle) dlclose(m->modhandle);
@@ -294,10 +338,11 @@ machine *parse_argument (char c, machine *m) {
 			char	**confv;
 
 			fstat(conf_fd, &sb);
-			conftext = malloc(sb.st_size);
+			conftext = malloc(sb.st_size+1);
 			/* pretend read always works, since this is from disk :-) */
 			read(conf_fd, conftext, sb.st_size);
 			close(conf_fd);
+			conftext[sb.st_size] = '\0';
 			
 			/* read the config file line by line */
 			for (confline = conftext;
@@ -334,6 +379,8 @@ int main(int argc, char **argv) {
 
 	machine_names = g_hash_table_new(g_str_hash, g_str_equal);
 	machine_fds = g_hash_table_new(g_int_hash, g_int_equal);
+	accept_fds = g_hash_table_new(g_int_hash, g_int_equal);
+	net_fds = g_hash_table_new(g_int_hash, g_int_equal);
 
 	if (getuid())
 		logpath = ".";
@@ -355,47 +402,122 @@ int main(int argc, char **argv) {
 	poptFreeContext(optCon);
 
 	/* set everything up -- daemon() after setup to report problems */
+	setup_logpath(logpath);
 	setup_pidfile();
-	setup_termios();
+	setup_devices();
 	if (do_daemon) {
 		daemon(1, 0);
 		setup_pidfile(); /* daemon() has changed pid ... */
 	}
 
 	/* set the signal handler to restore the old
-	 * termios handler */
-	sact.sa_handler = cleanup_termios_and_exit;
+	 * termios handler, etc. */
+	sact.sa_handler = cleanup_devices_and_exit;
 	sigaction(SIGINT, &sact, NULL);
-	sigaction(SIGPIPE, &sact, NULL);
 	sigaction(SIGTERM, &sact, NULL);
 	sact.sa_handler = reopen_logfiles;
 	sigaction(SIGHUP, &sact, NULL);
+	sact.sa_handler = SIG_IGN;
+	sigaction(SIGPIPE, &sact, NULL);
 
 	ufds = prepare_poll(&nfds);
 
 	do {
-		int i, n;
+	    int i, n, r;
 
-		poll(ufds, nfds, -1);
-		for (n = 0; n < nfds; n++) {
-			/* if (ufds[n].revents & POLLERR) FIXME: ??????? */
-			if (ufds[n].revents & POLLIN) {
-#				define BUFSIZE 1024
-				char    buf[BUFSIZE];
-				machine *m;
-				
-				m = g_hash_table_lookup(machine_fds, &ufds[n].fd);
-				if (!m) continue;
+	    r = poll(ufds, nfds, -1);
+	    if (r < 0) die("poll failed unexpectedly\n");
+	    for (n = 0; n < nfds; n++) {
+#		define TTYWATCH_BUFSIZE 1024
+		char buf[TTYWATCH_BUFSIZE];
+		machine *m;
+		net_conn *nc;
 
-                                i = read(m->dev_fd, buf, BUFSIZE-1);
-				if (i >= 1) {
-					buf[i+1] = '\0';
-					if (m->process_line)
-						m->process_line(m, buf);
-					else
-						write(m->log_fd, buf, i);
-				}
-			}
+		if (!ufds[n].revents) continue;
+
+		if ((ufds[n].revents & POLLHUP) || (ufds[n].revents & POLLERR)) {
+		    if ((nc = g_hash_table_lookup(net_fds, &ufds[n].fd))) {
+			remove_client (nc, ufds, &nfds);
+			break;
+		    } else {
+			/* If tcp monitored ports are later added, reopen
+			 * closed ports here
+			 */
+
+			/* should not happen, but if it does, we cannot recover;
+			 * we would go into an infinite poll loop instead.
+			 */
+			die("Unexpected HUP|ERR poll event %d on fd %d\n",
+			    ufds[n].revents, ufds[n].fd);
+		    }
 		}
+
+		if (ufds[n].revents & POLLIN) {
+		    /* is it incoming data from a port? */
+		    if ((m = g_hash_table_lookup(machine_fds, &ufds[n].fd))) {
+			i = read(m->dev_fd, buf, TTYWATCH_BUFSIZE-1);
+			if (i >= 1) {
+			    buf[i+1] = '\0';
+			    logfile_write(m->log_fd, buf, i);
+			    if (m->process_line)
+				m->process_line(m, buf);
+			    if (m->net_conns) {
+				GSList *l;
+
+				for (l = m->net_conns; l; l = g_slist_next(l)) {
+				    net_conn *nc = l->data;
+
+				    telnet_send_output(nc, buf, i);
+				}
+			    }
+			}
+
+		    /* is it an incoming connection request? */
+		    } else if ((m = g_hash_table_lookup(accept_fds, &ufds[n].fd))) {
+			int conn = socket_accept(ufds[n].fd);
+			if (conn) {
+			    net_conn *nc = calloc(sizeof(net_conn), 1);
+			    if (!nc) {
+				close(conn);
+				warn(m, "out of memory for new connection");
+			    }
+			    nc->m = m;
+			    nc->fd = conn;
+
+			    m->net_conns = g_slist_prepend(m->net_conns, nc);
+			    g_hash_table_insert(net_fds, &nc->fd, nc);
+			    telnet_negotiate(nc->fd);
+			    /* if available, send some existing logfile data */
+			    logfile_spew(nc);
+			    ufds = prepare_poll(&nfds); break;
+			}
+
+		    /* is it incoming typing on a client socket? */
+		    } else if ((nc = g_hash_table_lookup(net_fds, &ufds[n].fd))) {
+			m = nc->m;
+			i = read(ufds[n].fd, buf, TTYWATCH_BUFSIZE-1);
+			if (!i) {
+			    /* why not POLLHUP? <sigh> */
+			    remove_client (nc, ufds, &nfds);
+			    break;
+			}
+			i = telnet_process_input(nc, buf, i);
+			write(m->dev_fd, buf, i);
+			/* write(STDERR_FILENO, buf, i); */
+
+		    } else {
+			/* what are we missing? */
+			i = read(ufds[n].fd, buf, TTYWATCH_BUFSIZE-1);
+			warn(NULL, "error: throwing away %d bytes from unknown fd %d", i, ufds[n].fd);
+		    }
+
+		} else if (ufds[n].revents) {
+		    /* should not happen, but if it does, we cannot recover;
+		     * we would go into an infinite poll loop instead.
+		     */
+		    die("Unexpected poll event %d on fd %d\n",
+		        ufds[n].revents, ufds[n].fd);
+		}
+	    }
 	} while (1); /* exits through signal handler */
 }
