@@ -24,15 +24,38 @@
 
 #include <termios.h>
 #include <dlfcn.h>
+#include <glib.h>
+
+/* this should really be in telnet.h, but need to short-circuit
+ * the dependency
+ */
+typedef enum {
+	TS_DATA = 0,
+	TS_IAC,
+	TS_SB,
+	TS_SB_IAC,
+} telnet_state;
+
+typedef struct net_conn net_conn;
+struct net_conn {
+	struct machine *m;
+	int fd;
+	telnet_state ts;
+};
 
 typedef struct machine machine;
 struct machine {
 	char *name;
 	char *devpath;
+	char *ipport;
+	char *ipaddr;
+	int logspew;
 	void (*process_line)(machine *m, char *data);
 	void *modhandle;
 	int dev_fd;
 	int log_fd;
+	int accept_fd;
+	GSList *net_conns;
 	speed_t speed;
 	struct termios ots;
 	struct termios ts;
